package ru.amse.baltijsky.javascheme.importer.sablecc.java15;

import ru.amse.baltijsky.javascheme.importer.IImporter;
import ru.amse.baltijsky.javascheme.importer.sablecc.java15.analysis.DepthFirstAdapter;
import ru.amse.baltijsky.javascheme.importer.sablecc.java15.lexer.Lexer;
import ru.amse.baltijsky.javascheme.importer.sablecc.java15.lexer.LexerException;
import ru.amse.baltijsky.javascheme.importer.sablecc.java15.node.*;
import ru.amse.baltijsky.javascheme.importer.sablecc.java15.parser.Parser;
import ru.amse.baltijsky.javascheme.importer.sablecc.java15.parser.ParserException;
import ru.amse.baltijsky.javascheme.tree.*;
import static ru.amse.baltijsky.javascheme.tree.NodeType.*;
import ru.amse.baltijsky.javascheme.util.CharacterGroups;
import ru.amse.baltijsky.javascheme.util.Util;

import java.io.IOException;
import java.io.PushbackReader;
import java.io.Reader;
import java.util.*;
import java.util.regex.Pattern;

/**
 * A Java source code importer implemented via SableCC. Converts the given Java code into a schema tree.
 * Makes use of sablecc java 1.5 grammar with an If-Chaining feature intended to handle <code>IF-ELSE_IF-ELSE</code>
 * situations properly.
 * The reasoning behind implementing this feature is the fact that such constructions are parsed in this grammar as
 * <p/>
 * <code>IF-ELSE + [statement beginning with IF-ELSE + [statement beginning with IF-ELSE + [...]]]</code>
 * <p/>
 * To cope with it, we'll have two states: branching and non-branching. Branching is switched on every time a <tt>If</tt+<tt>Else</tt>
 * construction is encountered. When branching, all subsequent <tt>If</tt>s and <tt>IfElse</tt>s that come after Else are managed as
 * branches of the single first If. Also, any such branch-if does not create a separate level (contrary to a common If).
 * When a statement that goes after Else does not begin with a If, branching is switched off at the current level.
 * As nested if-else construction exist, multi-level branching can occur, thus we have to store all branch heads
 * for the levels where branching is currently switched on.
 */
@SuppressWarnings({"StringConcatenationInsideStringBufferAppend"})
public class SableCCImporter extends DepthFirstAdapter implements IImporter {
    /**
     * The task for the parsing.
     */
    private enum ParsingTask {
        BUILD_TREE,
        INSPECT_CLASSES,
        INSPECT_METHODS
    }

    /**
     * A pair of values for storing branch heads to enable multi-level branching.
     */
    private static class IfBranchHead {
        /**
         * The head of the branch in the schema tree.
         */
        public IfNode node;
        /**
         * Level of the direct descendants.
         */
        public int level;

        /**
         * Creates a new chain node with the specified level.
         *
         * @param theNode  the node
         * @param theLevel the level
         */
        public IfBranchHead(IfNode theNode, int theLevel) {
            node = theNode;
            level = theLevel;
        }
    }

    /**
     * Describes states of the parser with respect to the branching nodes. Used in branching.
     */
    private static enum BranchState {
        /**
         * nothing to do with branching, it may be switched on or off
         */
        NORMAL,
        // states to apply or promote branching
        // for the following states the first statement encountered must be attached to the current level branch head.
        /**
         * <tt>IF encountered and ELSE is known to follow later</tt>.
         */
        AFTER_IF,
        /**
         * <tt>ELSE IF</tt> encountered; <tt>ELSE</tt> is expected further in the level.
         */
        AFTER_ELSE_IF_ELSE,
        /**
         * <tt>ELSE IF</tt> encountered; no <tt>ELSE</tt> is expected further in the level.
         */
        AFTER_ELSE_IF_NOELSE,
        /**
         * <tt>ELSE</tt> encountered, and the first <tt>IF</tt> will continue branching.
         */
        AFTER_ELSE_BRANCHING,
        /**
         * <tt>ELSE</tt> encountered, any subsequent statement will terminate branching.
         */
        AFTER_ELSE_NOBRANCHING
    }

    /**
     * The parsing task.
     */
    private ParsingTask parsingTask = ParsingTask.BUILD_TREE;
    /**
     * Name of the class that will be imported.
     */
    private String className;
    /**
     * Indicates that the class to be imported was found.
     */
    private boolean classFound = false;
    /**
     * Name of the method that will be imported.
     */
    private String methodSignature;
    /**
     * Indicates that the method to be imported was found.
     */
    private boolean methodFound = false;
    /**
     * Currently processed node during the depth-first tree traversal.
     * Is used to link the newly created nodes and their parents or previous nodes as well.
     */
    private SchemaNode cur = null;
    /**
     * Current nesting level.
     */
    private int level = 0;
    /**
     * The last value of level when a new node was created (to calculate level delta).
     */
    private int lastLevel = 0;
    /**
     * Indicates whether the root node has been created already.
     */
    private boolean createdRoot = false;
    /**
     * The root of the constructed tree.
     */
    private SchemaNode root = null;
    /**
     * A stack to manage if chaining.
     */
    private Deque<IfBranchHead> ifBranchHeads = new LinkedList<IfBranchHead>();
    /**
     * Current state of branching.
     */
    private BranchState branchState = BranchState.NORMAL;
    /**
     * Enclosing statements for the current statement. Used to handle break and continue.
     */
    private Deque<SchemaNode> nestedLevels = new LinkedList<SchemaNode>();
    /**
     * Stack of <code>switch</code> constructions enclosing the current statement.
     */
    private Deque<SwitchNode> switches = new LinkedList<SwitchNode>();
    /**
     * Last registered label (will be added to the next encountered statement defining a nesting node).
     */
    private String label;
    /**
     * Repair schedule for <code>break</code> nodes. As we discover a nesting node out of which the break node leads,
     * we should set for this break node the following node of this nesting node as the node to jump. but the tree
     * isn't known at this point yet, and we store only the link to the mentioned nesting node. After the tree is built,
     * we'll take all such nodes and repair the links.
     */
    private List<BreakNode> breakSchedule = new ArrayList<BreakNode>();
    /**
     * List for inspection tasks.
     */
    private List<String> inspectionList = new ArrayList<String>();

    /**
     * The main loadTree function. Parses the input source into the internal representation tree and returns its root.
     *
     * @param in         the input source to loadTree
     * @param className  name of the class which method will be imported. May be empty, then the first encountered public class will be imported
     * @param methodName name of the method that will be imported
     */
    public SchemaTree loadTree(Reader in, String className, String methodName) throws IOException {
        parsingTask = ParsingTask.BUILD_TREE;
        this.className = className;
        this.methodSignature = methodName;
        setup();
        try {
            (new Parser(new Lexer(new PushbackReader(in)))).parse().apply(this);

            if (!classFound) {
                throw new IOException("failed to find the specified class " + className);
            } else if (!methodFound) {
                throw new IOException("failed to find the specified method " + methodName);
            }

            repairBreaksJumps();
            return new SchemaTree(root);
        } catch (ParserException e) {
            throw new IOException("parsing failed: " + e.getMessage() + " at " + e.getToken());
        } catch (LexerException e) {
            throw new IOException("parsing failed: " + e.getMessage());
        } catch (IllegalStateException e) {
            if (e.getCause() instanceof ParserException) {
                ParserException pe = (ParserException) e.getCause();
                throw new IOException("parsing failed: " + pe.getMessage() + " at " + pe.getToken());
            }
            throw e;
        }
    }

    /**
     * Returns the list of classes declared in the source code contained in the input reader.
     *
     * @param in the input reader
     * @return list of classes defined in the java source code contained in the input reader
     * @throws java.io.IOException if an error occurred while parsing the given input
     */
    public String[] getClasses(Reader in) throws IOException {
        parsingTask = ParsingTask.INSPECT_CLASSES;
        inspectionList.clear();
        setup();
        try {
            (new Parser(new Lexer(new PushbackReader(in)))).parse().apply(this);
            return getTrimmedInspectionArray();
        } catch (ParserException e) {
            throw new IOException("parsing failed: " + e.getMessage() + " at " + e.getToken());
        } catch (LexerException e) {
            throw new IOException("parsing failed: " + e.getMessage());
        } catch (IllegalStateException e) {
            if (e.getCause() instanceof ParserException) {
                ParserException pe = (ParserException) e.getCause();
                throw new IOException("parsing failed: " + pe.getMessage() + " at " + pe.getToken());
            }
            throw e;
        }
    }

    /**
     * Returns the list of methods declared in the specified class in the java source file contained in the input reader.
     *
     * @param in        the input reader
     * @param className the name of the class to collect methods
     * @return list of methods
     * @throws java.io.IOException if an error occurred while parsing the given input or the class was not found
     */
    public String[] getMethods(Reader in, String className) throws IOException {
        parsingTask = ParsingTask.INSPECT_METHODS;
        inspectionList.clear();
        this.className = className;
        setup();
        try {
            (new Parser(new Lexer(new PushbackReader(in)))).parse().apply(this);
            if (!classFound) {
                throw new IOException("failed to find the specified class " + className);
            }
            return getTrimmedInspectionArray();
        } catch (ParserException e) {
            throw new IOException("parsing failed: " + e.getMessage() + " at " + e.getToken());
        } catch (LexerException e) {
            throw new IOException("parsing failed: " + e.getMessage());
        } catch (IllegalStateException e) {
            if (e.getCause() instanceof ParserException) {
                ParserException pe = (ParserException) e.getCause();
                throw new IOException("parsing failed: " + pe.getMessage() + " at " + pe.getToken());
            }
            throw e;
        }
    }

    /**
     * Returns inspection list converted to array of string with trimmed whitespaces in the beginning and the end.
     *
     * @return trimmed inspection array
     */
    private String[] getTrimmedInspectionArray() {
        String[] array = inspectionList.toArray(new String[inspectionList.size()]);
        for (int i = 0; i < array.length; i++) {
            array[i] = array[i].trim();
        }
        return array;
    }

    /**
     * Setup before parsing common to all tasks.
     */
    private void setup() {
        classFound = false;
        methodFound = false;
        level = 0;
        lastLevel = 0;
        createdRoot = false;
        ifBranchHeads.clear();
        nestedLevels.clear();
        switches.clear();
        breakSchedule.clear();
        branchState = BranchState.NORMAL;
    }


    /**
     * Adjusts jump links for <code>break</code> nodes.
     */
    private void repairBreaksJumps() {
        for (BreakNode b : breakSchedule) {
            SchemaNode jumpNode = b.getNodeToJump();
            while (jumpNode != null && !jumpNode.hasNext()) {
                jumpNode = jumpNode.getParent();
            }
            if (jumpNode != null) {
                // for break outside of try, we must go to finally if there is one
                if (jumpNode instanceof TryNode && ((TryNode) jumpNode).isHasFinally()) {
                    jumpNode = ((TryNode) jumpNode).getFinallyNode();
                } else {
                    jumpNode = jumpNode.getNext();
                }
            }
            b.setNodeToJump(jumpNode);
        }
    }

    @Override
    public void inACompilationUnit(ACompilationUnit node) {
        cur = createNode(TRANSLATION_UNIT, adjustNodeText(node.toString()));
        increaseLevel();
    }

    @Override
    public void outACompilationUnit(ACompilationUnit node) {
        decreaseLevel();
    }

    @Override
    public void inAClassDeclaration(AClassDeclaration node) {
        cur = createNode(CLASS, adjustNodeText(node.getIdentifier().toString()));
        increaseLevel();
    }

    @Override
    public void outAClassDeclaration(AClassDeclaration node) {
        decreaseLevel();
    }

    @Override
    public void caseAClassDeclaration(AClassDeclaration node) {
        if (parsingTask == ParsingTask.INSPECT_CLASSES) {
            inspectionList.add(node.getIdentifier().toString());
        } else {
            boolean skip = classFound || !Util.equalIgnoreWhitespace(node.getIdentifier().toString(), className);
            // for an empty classname, check whether the class is public
            if (className.equals("")) {
                for (PModifier mod : node.getModifier()) {
                    if (Util.equalIgnoreWhitespace(mod.toString(), "public")) {
                        skip = false;
                        break;
                    }
                }
            }
            if (!skip) {
                classFound = true;
                super.caseAClassDeclaration(node);
            }
        }
    }

    @Override
    public void inATypeMethodDeclaration(ATypeMethodDeclaration node) {
        // obtain method header
        StringBuilder methodHeader = new StringBuilder();
        if (node.getModifier() != null) {
            boolean first = true;
            for (PModifier modifier : node.getModifier()) {
                methodHeader.append(first ? "" : ", ").append(modifier.toString().trim());
                first = false;
            }
            methodHeader.append(node.getModifier().toString().trim());
        }
        if (node.getTypeParameters() != null) {
            methodHeader.append(node.getTypeParameters());
        }
        if (node.getType() != null) {
            methodHeader.append(" " + node.getType());
        }
        if (node.getDim() != null) {
            methodHeader.append(node.getDim());
        }
        if (node.getIdentifier() != null) {
            methodHeader.append(node.getIdentifier());
        }
        if (node.getFormalParameters() != null) {
            methodHeader.append(new PrettyPrinter().retrieveParameters(node.getFormalParameters()));
        }
        if (node.getThrowsException() != null) {
            methodHeader.append(node.getThrowsException().toString().trim());
        }
        cur = createNode(METHOD, adjustNodeText(methodHeader.toString()));
        increaseLevel();
    }

    @Override
    public void outATypeMethodDeclaration(ATypeMethodDeclaration node) {
        decreaseLevel();
    }

    @Override
    public void caseATypeMethodDeclaration(ATypeMethodDeclaration node) {
        String signature = new PrettyPrinter().retrieveSignature(node.getIdentifier().toString(), node.getFormalParameters());
        if (parsingTask == ParsingTask.INSPECT_METHODS) {
            inspectionList.add(signature);
        } else if (!methodFound && Util.equalIgnoreWhitespace(signature, methodSignature)) {
            methodFound = true;
            super.caseATypeMethodDeclaration(node);
        }
    }

    public void inAVoidMethodDeclaration(AVoidMethodDeclaration node) {
        // obtain method header
        StringBuilder methodHeader = new StringBuilder();
        if (node.getModifier() != null) {
            methodHeader.append(node.getModifier().toString().trim());
        }
        if (node.getModifier() != null) {
            boolean first = true;
            for (PModifier modifier : node.getModifier()) {
                methodHeader.append(first ? "" : ", ").append(modifier.toString().trim());
                first = false;
            }
            methodHeader.append(node.getModifier().toString().trim());
        }
        if (node.getVoid() != null) {
            methodHeader.append(" " + node.getVoid());
        }
        if (node.getIdentifier() != null) {
            methodHeader.append(node.getIdentifier());
        }
        if (node.getFormalParameters() != null) {
            methodHeader.append(new PrettyPrinter().retrieveParameters(node.getFormalParameters()));
        }
        if (node.getThrowsException() != null) {
            methodHeader.append(node.getThrowsException().toString().trim());
        }
        cur = createNode(METHOD, adjustNodeText(methodHeader.toString()));
        increaseLevel();
    }

    public void outAVoidMethodDeclaration(AVoidMethodDeclaration node) {
        decreaseLevel();
    }

    @Override
    public void caseAVoidMethodDeclaration(AVoidMethodDeclaration node) {
        String signature = new PrettyPrinter().retrieveSignature(node.getIdentifier().toString(), node.getFormalParameters());
        if (parsingTask == ParsingTask.INSPECT_METHODS) {
            inspectionList.add(signature);
        } else if (!methodFound && Util.equalIgnoreWhitespace(signature, methodSignature)) {
            methodFound = true;
            super.caseAVoidMethodDeclaration(node);
        }
    }

    @Override
    public void inAConstructorDeclaration(AConstructorDeclaration node) {
        // obtain constructor header
        StringBuilder ctorHeader = new StringBuilder();
        if (node.getModifier() != null) {
            boolean first = true;
            for (PModifier modifier : node.getModifier()) {
                ctorHeader.append(first ? "" : ", ").append(modifier.toString().trim());
                first = false;
            }
            ctorHeader.append(node.getModifier().toString().trim());
        }
        if (node.getIdentifier() != null) {
            ctorHeader.append(node.getIdentifier());
        }
        if (node.getFormalParameters() != null) {
            ctorHeader.append(new PrettyPrinter().retrieveParameters(node.getFormalParameters()));
        }
        if (node.getThrowsException() != null) {
            ctorHeader.append(node.getThrowsException().toString());
        }
        cur = createNode(METHOD, adjustNodeText(ctorHeader.toString()));
        increaseLevel();
    }

    @Override
    public void outAConstructorDeclaration(AConstructorDeclaration node) {
        decreaseLevel();
    }

    @Override
    public void caseAConstructorDeclaration(AConstructorDeclaration node) {
        String signature = new PrettyPrinter().retrieveSignature(node.getIdentifier().toString(), node.getFormalParameters());
        if (parsingTask == ParsingTask.INSPECT_METHODS) {
            inspectionList.add(signature);
        } else if (!methodFound && Util.equalIgnoreWhitespace(signature, methodSignature)) {
            methodFound = true;
            super.caseAConstructorDeclaration(node);
        }
    }

    @Override
    public void inALabeledStatement(ALabeledStatement node) {
        label = node.getIdentifier().toString();
    }

    @Override
    public void inALabeledStatementNoShortIf(ALabeledStatementNoShortIf node) {
        label = node.getIdentifier().toString();
    }

    @Override
    public void inAWhileStatement(AWhileStatement node) {
        cur = createNode(WHILE, adjustNodeText(getExpression(node.getCondition())));
        increaseLevel(true);
    }

    @Override
    public void outAWhileStatement(AWhileStatement node) {
        decreaseLevel(true);
    }

    @Override
    public void inADoStatement(ADoStatement node) {
        cur = createNode(DO_WHILE, adjustNodeText(node.getCondition().toString()));
        increaseLevel(true);
    }

    @Override
    public void outADoStatement(ADoStatement node) {
        decreaseLevel(true);
    }

    @Override
    public void inAForStatement(AForStatement node) {
        String forInit = node.getForInit() != null ? node.getForInit().toString() : "";
        String expression = node.getExpression() != null ? node.getExpression().toString() : "";
        String forUpdate = node.getForUpdate() != null ? node.getForUpdate().toString() : "";
        cur = createNode(FOR, adjustNodeText(forInit + "; " + expression + "; " + forUpdate));
        increaseLevel(true);
    }

    @Override
    public void outAForStatement(AForStatement node) {
        decreaseLevel(true);
    }

    @Override
    public void caseAForStatement(AForStatement node) {
        inAForStatement(node);
        if (node.getFor() != null) {
            node.getFor().apply(this);
        }
        if (node.getLPar() != null) {
            node.getLPar().apply(this);
        }
        // we don't visit ForInit to omit adding this action to the program flow
        if (node.getSemicolon1() != null) {
            node.getSemicolon1().apply(this);
        }
        if (node.getExpression() != null) {
            node.getExpression().apply(this);
        }
        if (node.getSemicolon2() != null) {
            node.getSemicolon2().apply(this);
        }
        if (node.getForUpdate() != null) {
            node.getForUpdate().apply(this);
        }
        if (node.getRPar() != null) {
            node.getRPar().apply(this);
        }
        if (node.getStatement() != null) {
            node.getStatement().apply(this);
        }
        outAForStatement(node);
    }

    public void inAForStatementNoShortIf(AForStatementNoShortIf node) {
        String forInit = node.getForInit() != null ? node.getForInit().toString() : "";
        String expression = node.getExpression() != null ? node.getExpression().toString() : "";
        String forUpdate = node.getForUpdate() != null ? node.getForUpdate().toString() : "";
        cur = createNode(FOR, adjustNodeText(forInit + "; " + expression + "; " + forUpdate));
        increaseLevel(true);
    }

    public void outAForStatementNoShortIf(AForStatementNoShortIf node) {
        decreaseLevel(true);
    }

    @Override
    public void caseAForStatementNoShortIf(AForStatementNoShortIf node) {
        inAForStatementNoShortIf(node);
        if (node.getFor() != null) {
            node.getFor().apply(this);
        }
        // we don't visit ForInit to omit adding this action to the program flow
        if (node.getExpression() != null) {
            node.getExpression().apply(this);
        }
        if (node.getForUpdate() != null) {
            node.getForUpdate().apply(this);
        }
        if (node.getStatementNoShortIf() != null) {
            node.getStatementNoShortIf().apply(this);
        }
        outAForStatementNoShortIf(node);
    }

    public void inAEnhancedForStatement(AEnhancedForStatement node) {
        StringBuilder mod = new StringBuilder();
        if (node.getModifier() != null) {
            boolean first = true;
            for (PModifier modifier : node.getModifier()) {
                mod.append(first ? "" : ", ").append(modifier.toString());
                first = false;
            }
        }
        String type = node.getType() != null ? node.getType().toString() : "";
        String id = node.getIdentifier() != null ? node.getIdentifier().toString() : "";
        String expr = node.getIdentifier() != null ? node.getExpression().toString() : "";
        cur = createNode(FOR, adjustNodeText(mod + type + id + " : " + expr));
        increaseLevel(true);
    }

    public void outAEnhancedForStatement(AEnhancedForStatement node) {
        decreaseLevel(false);
    }

    @Override
    public void caseAEnhancedForStatement(AEnhancedForStatement node) {
        inAEnhancedForStatement(node);
        if (node.getStatement() != null) {
            node.getStatement().apply(this);
        }
        outAEnhancedForStatement(node);
    }

    public void inAEnhancedForStatementNoShortIf(AEnhancedForStatementNoShortIf node) {
        StringBuilder mod = new StringBuilder();
        if (node.getModifier() != null) {
            boolean first = true;
            for (PModifier modifier : node.getModifier()) {
                mod.append(first ? "" : ", ").append(modifier.toString());
                first = false;
            }
        }
        String type = node.getType() != null ? node.getType().toString() : "";
        String id = node.getIdentifier() != null ? node.getIdentifier().toString() : "";
        String expr = node.getIdentifier() != null ? node.getExpression().toString() : "";
        cur = createNode(FOR, adjustNodeText(mod + " " + type + " " + id + " : " + expr));
        increaseLevel(true);
    }

    public void outAEnhancedForStatementNoShortIf(AEnhancedForStatementNoShortIf node) {
        decreaseLevel(false);
    }

    @Override
    public void caseAEnhancedForStatementNoShortIf(AEnhancedForStatementNoShortIf node) {
        inAEnhancedForStatementNoShortIf(node);
        if (node.getStatementNoShortIf() != null) {
            node.getStatementNoShortIf().apply(this);
        }
        outAEnhancedForStatementNoShortIf(node);
    }

    @Override
    public void inAExpressionStatement(AExpressionStatement node) {
        cur = createNode(ACTION, adjustNodeText(node.toString()));
    }

    @Override
    public void inALocalVariableDeclarationStatement(ALocalVariableDeclarationStatement node) {
        cur = createNode(ACTION, adjustNodeText(node.toString()));
    }

    @Override
    public void inAThisExplicitConstructorInvocation(AThisExplicitConstructorInvocation node) {
        cur = createNode(ACTION, adjustNodeText(node.toString()));
    }

    @Override
    public void inASuperExplicitConstructorInvocation(ASuperExplicitConstructorInvocation node) {
        cur = createNode(ACTION, adjustNodeText(node.toString()));
    }

    @Override
    public void inAQualifiedExplicitConstructorInvocation(AQualifiedExplicitConstructorInvocation node) {
        cur = createNode(ACTION, adjustNodeText(node.toString()));
    }

    @Override
    public void inAEmptyStatement(AEmptyStatement node) {
        cur = createNode(ACTION, adjustNodeText("// empty statement"));
    }

    @Override
    public void inAAssertStatementShort(AAssertStatementShort node) {
        cur = createNode(ACTION, adjustNodeText(node.toString()));
    }

    @Override
    public void inAAssertStatementLong(AAssertStatementLong node) {
        cur = createNode(ACTION, adjustNodeText(node.toString()));
    }

    @Override
    public void inASynchronizedStatement(ASynchronizedStatement node) {
        cur = createNode(SYNCHRONIZED, adjustNodeText(getExpression(node.getCondition())));
        increaseLevel(true);
    }

    @Override
    public void outASynchronizedStatement(ASynchronizedStatement node) {
        decreaseLevel(true);
    }

    // JUMP constructions
    //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void inABreakStatement(ABreakStatement node) {
        cur = createNode(BREAK, adjustNodeText(node.toString()));
        SchemaNode enclosingNode = null;
        if (node.getIdentifier() != null) {
            String label = node.getIdentifier().toString();
            for (SchemaNode nestedLevel : nestedLevels) {
                enclosingNode = nestedLevel;
                if (Util.equalIgnoreWhitespace(enclosingNode.getLabel(), label)) {
                    break;
                }
            }
            if (enclosingNode == null) {
                throw new IllegalStateException(new ParserException(new TBreak(), "Undefined label: " + label));
            }
        } else {
            for (Iterator<SchemaNode> i = nestedLevels.descendingIterator(); i.hasNext();) {
                enclosingNode = i.next();
                if (enclosingNode instanceof ILoopNode || enclosingNode instanceof SwitchNode) {
                    break;
                }
            }
            if (enclosingNode == null) {
                throw new IllegalStateException(new ParserException(new TBreak(), "break outside switch or loop"));
            }
        }
        // we should set the node immediately after the found nestingNode,
        // but the tree is actually not built there. we'll repair the links after the tree is built,
        // so we have to add cur to the repair schedule.
        ((FlowChangeNode) cur).setNodeToJump(enclosingNode);
        breakSchedule.add((BreakNode) cur);

    }

    @Override
    public void inAContinueStatement(AContinueStatement node) {
        cur = createNode(CONTINUE, adjustNodeText(node.toString()));
        SchemaNode enclosingNode = null;
        for (Iterator<SchemaNode> i = nestedLevels.descendingIterator(); i.hasNext();) {
            enclosingNode = i.next();
            if (enclosingNode instanceof ILoopNode || enclosingNode instanceof SwitchNode) {
                break;
            }
        }
        if (enclosingNode == null) {
            throw new IllegalStateException(new ParserException(new TContinue(), "continue outside of loop"));
        }
        ((FlowChangeNode) cur).setNodeToJump(enclosingNode);
    }

    @Override
    public void inAReturnStatement(AReturnStatement node) {
        cur = createNode(RETURN, adjustNodeText(node.toString()));
    }

    @Override
    public void inAThrowStatement(AThrowStatement node) {
        cur = createNode(THROW, adjustNodeText(node.toString()));
    }

    // TRY constructions
    //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void inATryTryStatement(ATryTryStatement node) {
        cur = createNode(TRY, adjustNodeText("try"));
        TryNode tryNode = (TryNode) cur;
        tryNode.setHasCatchClauses(true);
        tryNode.setHasFinally(false);
        increaseLevel(true);
    }

    @Override
    public void outATryTryStatement(ATryTryStatement node) {
        decreaseLevel(true);
    }

    @Override
    public void caseATryTryStatement(ATryTryStatement node) {
        inATryTryStatement(node);
        if (node.getTry() != null) {
            node.getTry().apply(this);
        }
        if (node.getBlock() != null) {
            node.getBlock().apply(this);
        }
        {
            decreaseLevel(false);
            cur = createNode(CATCH, adjustNodeText(""));
            CatchNode catchNode = (CatchNode) cur;
            catchNode.setHasFinally(false);
            List<PCatchClause> copy = new ArrayList<PCatchClause>(node.getCatchClause());
            for (PCatchClause e : copy) {
                catchNode.addConditionToNewSet(adjustText(((ACatchClause) e).getFormalParameters().toString()));
                lastLevel = level;
                increaseLevel(false);
                e.apply(this);
                decreaseLevel(false);
            }
            increaseLevel(false);
        }
        outATryTryStatement(node);
    }

    @Override
    public void inAFinallyTryStatement(AFinallyTryStatement node) {
        super.inAFinallyTryStatement(node);
    }

    @Override
    public void outAFinallyTryStatement(AFinallyTryStatement node) {
        decreaseLevel(true);
    }

    @Override
    public void caseAFinallyTryStatement(AFinallyTryStatement node) {
        cur = createNode(TRY, adjustNodeText("try"));
        TryNode tryNode = (TryNode) cur;
        tryNode.setHasFinally(true);
        increaseLevel(true);
        inAFinallyTryStatement(node);

        if (node.getTry() != null) {
            node.getTry().apply(this);
        }
        if (node.getBlock() != null) {
            node.getBlock().apply(this);
        }
        CatchNode catchNode;
        {
            decreaseLevel(false);
            cur = createNode(CATCH, adjustNodeText(""));
            catchNode = (CatchNode) cur;
            List<PCatchClause> copy = new ArrayList<PCatchClause>(node.getCatchClause());
            tryNode.setHasCatchClauses(!copy.isEmpty());
            for (PCatchClause e : copy) {
                catchNode.addConditionToNewSet(adjustText(((ACatchClause) e).getFormalParameters().toString()));
                lastLevel = level;
                increaseLevel(false);
                e.apply(this);
                decreaseLevel(false);
            }
            increaseLevel(false);
        }
        if (node.getFinallyClause() != null) {
            if (catchNode != null) {
                catchNode.setHasFinally(true);
            }
            decreaseLevel(false);
            node.getFinallyClause().apply(this);
            increaseLevel(false);
        } else if (catchNode != null) {
            catchNode.setHasFinally(false);
        }
        outAFinallyTryStatement(node);
    }

    @Override
    public void inAFinallyClause(AFinallyClause node) {
        cur = createNode(FINALLY, adjustNodeText(""));
        label = nestedLevels.peek().getLabel();
        increaseLevel(true);
    }

    @Override
    public void outAFinallyClause(AFinallyClause node) {
        decreaseLevel(true);
    }

    // SWITCH constructions
    //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void inASwitchStatement(ASwitchStatement node) {
        cur = createNode(SWITCH, adjustNodeText("switch (" + getExpression(node.getCondition()) + ")"));
        switches.push((SwitchNode) cur);
        increaseLevel(true);
    }

    @Override
    public void outASwitchStatement(ASwitchStatement node) {
        switches.pop();
        decreaseLevel(true);
    }

    @Override
    public void caseASwitchBlockStatementGroup(ASwitchBlockStatementGroup node) {
        inASwitchBlockStatementGroup(node);
        {
            List<PSwitchLabel> copy = new ArrayList<PSwitchLabel>(node.getSwitchLabel());
            SwitchNode lastSwitch = switches.peek();
            lastSwitch.addNewConditionSet();
            for (PSwitchLabel e : copy) {
                lastSwitch.addConditionToCurrentSet(adjustText(e.toString()));
            }
        }
        {
            List<PBlockStatement> copy = new ArrayList<PBlockStatement>(node.getBlockStatement());
            lastLevel = level;
            increaseLevel(false);
            cur = nestedLevels.peek();
            for (PBlockStatement e : copy) {
                e.apply(this);
            }
            decreaseLevel(false);
        }
        outASwitchBlockStatementGroup(node);

    }

    // IF constructions
    //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Determines whether an <code>IF</code> statement in the current branching state should be treated as a
     * nested construction.
     *
     * @return true when <code>IF</code> should be treated as a nested construction
     */
    private boolean isNestedIf() {
        return branchState == BranchState.AFTER_IF
                || branchState == BranchState.AFTER_ELSE_IF_NOELSE
                || branchState == BranchState.AFTER_ELSE_IF_ELSE
                || branchState == BranchState.AFTER_ELSE_NOBRANCHING;
    }

    /**
     * Determines whether an <code>IF</code> statement in the current branching state should take part in branching.
     *
     * @return true when <code>IF</code> participates in branching
     */
    private boolean isBranchingIf() {
        return branchState == BranchState.AFTER_ELSE_BRANCHING;
    }

    /**
     * Takes the decision whether to promote or stop branching at the current level and returns the head of the current
     * (closed or unclosed) branching level.
     *
     * @return head of the current (closed or unclosed) branching level
     */
    private IfNode deduceBranchingTactics() {
        IfNode ifHead;
        if (branchState == BranchState.AFTER_ELSE_IF_ELSE || branchState == BranchState.AFTER_IF) {
            ifHead = ifBranchHeads.peek().node;
        } else {
            // stop branching here
            ifHead = ifBranchHeads.pop().node;
            if (branchState == BranchState.AFTER_ELSE_NOBRANCHING || branchState == BranchState.AFTER_ELSE_BRANCHING) {
                ifHead.addConditionToNewSet("else");
            }
        }
        return ifHead;
    }

    @Override
    public void caseAIfThenStatement(AIfThenStatement node) {
        boolean createdNestedLevel = false;
        // if we were branching already, here's the place where we switch it off (because we've no Else at this level)
        if (isBranchingIf()) {
            branchState = BranchState.AFTER_ELSE_IF_NOELSE; // no ELSE is expected to happen
            // switch branching off and add the condition
            ifBranchHeads.peek().node.addConditionToNewSet(adjustText(getExpression(node.getCondition())));
        } else {
            if (isNestedIf()) {
                // apply branching as if it were a simple node
                IfNode ifHead = deduceBranchingTactics();
                branchState = BranchState.NORMAL;
                cur = SchemaNode.createNode(IF, ifHead, null, adjustNodeText("if"));
                ifHead.addChild(cur);
            } else {
                cur = createNode(IF, adjustNodeText("if"));
            }
            // it's an If with a single statement, no branching needed; handle it like a common node with a child

            ((IfNode) cur).addConditionToNewSet(adjustText(getExpression(node.getCondition())));
            lastLevel = level;
            increaseLevel(true);
            createdNestedLevel = true;
        }

        // do the common visit
        super.caseAIfThenStatement(node);

        // set the level back if we've created it
        if (createdNestedLevel) {
            decreaseLevel(true);
        }
    }

    @Override
    public void caseAIfThenElseStatement(AIfThenElseStatement node) {
        // a flag that tells whether we've created a nested level while processing the node
        boolean createdNestedLevel = false;

        if (isBranchingIf()) {
            branchState = BranchState.NORMAL;
            ifBranchHeads.peek().node.addConditionToNewSet(adjustText(getExpression(node.getCondition())));
        } else {
            if (isNestedIf()) {
                // apply branching as if it were a simple node
                IfNode ifHead = deduceBranchingTactics();
                branchState = BranchState.NORMAL;
                // and start a new if-construction attached to the head
                cur = SchemaNode.createNode(IF, ifHead, null, adjustNodeText("if"));
                ifHead.addChild(cur);
            } else {
                // create an independent if-construction
                cur = createNode(IF, adjustNodeText("if"));
            }

            ((IfNode) cur).addConditionToNewSet(adjustText(getExpression(node.getCondition())));
            lastLevel = level;
            increaseLevel(true);
            createdNestedLevel = true;
            // create a new level of branching
            ifBranchHeads.push(new IfBranchHead((IfNode) cur, level));
        }

        inAIfThenElseStatement(node);
        if (node.getIf() != null) {
            node.getIf().apply(this);
        }
        if (node.getCondition() != null) {
            node.getCondition().apply(this);
        }
        if (node.getIfs() != null) {
            if (createdNestedLevel) {
                branchState = BranchState.AFTER_IF;
            } else if (node.getElses() != null) {
                branchState = BranchState.AFTER_ELSE_IF_ELSE;
            } else {
                branchState = BranchState.AFTER_ELSE_IF_NOELSE;
            }
            node.getIfs().apply(this);
        }
        if (node.getElse() != null) {
            node.getElse().apply(this);
        }
        if (node.getElses() != null) {
            branchState = BranchState.AFTER_ELSE_BRANCHING;
            node.getElses().apply(this);
        }
        outAIfThenElseStatement(node);

        // if we created nesting level, close it
        if (createdNestedLevel) {
            decreaseLevel(true);
        }
    }

    @Override
    public void caseAIfThenElseStatementNoShortIf(AIfThenElseStatementNoShortIf node) {
        // we may be here when after Else (= branching is on) or after ElseIf (= we handle it as a simple If without branching)

        // a flag that tells whether we've created a nested level while processing the node
        boolean createdNestedLevel = false;

        if (isBranchingIf()) {
            branchState = BranchState.NORMAL;
            ifBranchHeads.peek().node.addConditionToNewSet(adjustText(getExpression(node.getCondition())));
        } else {
            if (isNestedIf()) {
                // apply branching as if it were a simple node
                IfNode ifHead = deduceBranchingTactics();
                branchState = BranchState.NORMAL;
                // and start a new if-construction attached to the head
                cur = SchemaNode.createNode(IF, ifHead, null, adjustNodeText("if"));
                ifHead.addChild(cur);
            } else {
                // create an independent if-construction
                cur = createNode(IF, adjustNodeText("if"));
            }

            ((IfNode) cur).addConditionToNewSet(adjustText(getExpression(node.getCondition())));
            lastLevel = level;
            increaseLevel(true);
            createdNestedLevel = true;
            // create a new level of branching
            ifBranchHeads.push(new IfBranchHead((IfNode) cur, level));
        }

        inAIfThenElseStatementNoShortIf(node);
        if (node.getIf() != null) {
            node.getIf().apply(this);
        }
        if (node.getCondition() != null) {
            node.getCondition().apply(this);
        }
        if (node.getIfs() != null) {
            if (createdNestedLevel) {
                branchState = BranchState.AFTER_IF;
            } else if (node.getElses() != null) {
                branchState = BranchState.AFTER_ELSE_IF_ELSE;
            } else {
                branchState = BranchState.AFTER_ELSE_IF_NOELSE;
            }
            node.getIfs().apply(this);
        }
        if (node.getElse() != null) {
            node.getElse().apply(this);
        }
        if (node.getElses() != null) {
            branchState = BranchState.AFTER_ELSE_BRANCHING;
            node.getElses().apply(this);
        }
        outAIfThenElseStatementNoShortIf(node);

        if (createdNestedLevel) {
            decreaseLevel(true);
        }
    }

    // Needed to adjust the branching mechanism of defining the relationship after IF/ELSE IF/ELSE
    @Override
    public void inABlock(ABlock node) {
        if (branchState == BranchState.AFTER_ELSE_BRANCHING) {
            // if a brace comes after ELSE, there will be no construction such as ELSE IF, so a subsequent IF
            // should start its own branching.
            branchState = BranchState.AFTER_ELSE_NOBRANCHING;
        }
        if (node.getBlockStatement() == null || node.getBlockStatement().size() == 0) {
            // empty block: {}
            // create empty action
            cur = createNode(ACTION, adjustNodeText("// empty block"));
        } else {

        }
    }

    // Helper functions.
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Takes the text intended to be used as the node text and deletes all whitespaces between punctuation and non-punctuation,
     * returning the array with the sole element being the adjusted node text.
     *
     * @param text desired node text
     * @return array with the sole element being the adjusted node text
     */
    private String[] adjustNodeText(String text) {
        return new String[]{adjustText(text)};
    }

    /**
     * Adjusts the given java source text so that there're no whitespaces before and after some punctuation symbols.
     *
     * @param text the text to adjust
     * @return adjusted text
     */
    private String adjustText(String text) {
        String punctWhiteDisallowedAfter = "([!\"#$'().@_`~\\[\\]])";
        String punctWhiteDisallowedBefore = "([!\"#$'().@_`~\\[\\]:,;])";
        String letter = "(" + CharacterGroups.JAVA_LETTER_OR_DIGIT_GROUP + ")";
        String s = text;
        s = removeRedundantWhitespacesByPattern(s, letter + "[\\s]+" + punctWhiteDisallowedBefore);
        s = removeRedundantWhitespacesByPattern(s, punctWhiteDisallowedAfter + "[\\s]+" + letter);
        s = removeRedundantWhitespacesByPattern(s, punctWhiteDisallowedAfter + "[\\s]+" + punctWhiteDisallowedBefore);
        return s.trim();
    }

    private String removeRedundantWhitespacesByPattern(String s, String regex) {
        Pattern pattern = Pattern.compile(regex);
        while (pattern.matcher(s).find()) {
            s = pattern.matcher(s).replaceAll("$1$2");
        }
        return s;
    }

    /**
     * Extracts expression from condition.
     *
     * @param cond condition
     * @return string representing the expression in the condition
     */
    private String getExpression(PCondition cond) {
        return ((ACondition) cond).getExpression().toString();
    }

    /**
     * Creates a new node of the specified type and attaches it to the current node judging by the current level delta
     * (the change of level between the current and the newly created node).
     *
     * @param type the type of the node to create
     * @param s    code of the node
     * @return the created node
     */
    private SchemaNode createNode(NodeType type, String[] s) {
        if (!createdRoot) {
            createdRoot = true;
            root = SchemaNode.createNode(type, null, null, s);
            return root;
        }

        SchemaNode ret;

        // apply branching if necessary
        if (branchState != BranchState.NORMAL) {
            IfNode head = deduceBranchingTactics();
            branchState = BranchState.NORMAL;

            ret = SchemaNode.createNode(type, head, null, s);
            head.addChild(ret);
        } else {
            if (level - lastLevel > 0) {
                // create node at a deeper level
                ret = SchemaNode.createNode(type, cur, null, s);
                cur.addChild(ret);
            } else if (level - lastLevel == 0) {
                // create node at the same level
                ret = SchemaNode.createNode(type, cur.getParent(), cur, s);
            } else {
                // create a node at a shallower level
                SchemaNode p = cur;
                for (int i = level - lastLevel; i < 0; ++i) {
                    p = p.getParent();
                }
                ret = SchemaNode.createNode(type, p != null ? p.getParent() : p, p, s);
            }
        }
        lastLevel = level;
        return ret;
    }

    /**
     * Crates a new nested level.
     *
     * @param store determines whether to store the current level to be d in the stack
     */
    private void increaseLevel(boolean store) {
        ++level;
        if (store) {
            nestedLevels.push(cur);
            cur.setLabel(label);
            label = "";
        }
    }

    private void increaseLevel() {
        increaseLevel(false);
    }

    /**
     * Rejects the last nested level.
     *
     * @param discard reove current level from the stack
     */
    private void decreaseLevel(boolean discard) {
        --level;
        if (discard) {
            nestedLevels.pop();
        }
    }

    private void decreaseLevel() {
        decreaseLevel(false);
    }

    public SchemaTree[] loadTrees(Reader in) throws IOException {
        return new SchemaTree[0];
    }
}
